@@ -1,7 +1,7 @@
 package Complete::Getopt::Long;
 
-our $DATE = '2014-07-24'; # DATE
-our $VERSION = '0.03'; # VERSION
+our $DATE = '2014-07-26'; # DATE
+our $VERSION = '0.04'; # VERSION
 
 use 5.010001;
 use strict;
@@ -15,6 +15,33 @@ our @EXPORT_OK = qw(
 
 our %SPEC;
 
+# return the key/element if $opt expands unambiguously to exactly one
+# key/element in $opts (which can be a hash or array), otherwise return undef.
+# e.g. _expand1('--fo', [qw/--foo --bar --baz/]) is true, but _expand1('--ba',
+# ...) or _expand1('--qux', ...) are undef.
+sub _expand1 {
+    my ($opt, $opts) = @_;
+    my @candidates;
+    my $is_hash = ref($opts) eq 'HASH';
+    for ($is_hash ? (keys %$opts) : @$opts) {
+        next unless index($_, $opt) == 0;
+        push @candidates, $is_hash ? $opts->{$_} : $_;
+    }
+    return @candidates == 1 ? $candidates[0] : undef;
+}
+
+# mark an option (and all its aliases) as seen
+sub _mark_seen {
+    my ($seen_opts, $opt, $opts) = @_;
+    my $opthash = $opts->{$opt};
+    return unless $opthash;
+    my $ospec = $opthash->{ospec};
+    for (keys %$opts) {
+        my $v = $opts->{$_};
+        $seen_opts->{$_}++ if $v->{ospec} eq $ospec;
+    }
+}
+
 $SPEC{complete_cli_arg} = {
     v => 1.1,
     summary => 'Complete command-line argument using '.
@@ -62,7 +89,7 @@ _
             schema      => 'hash*',
         },
         words => {
-            summary     => 'Command line, already broken into words',
+            summary     => 'Command line arguments, like @ARGV',
             description => <<'_',
 
 See function `parse_cmdline` in `Complete::Bash` on how to produce this (if
@@ -95,29 +122,6 @@ the result of this function for bash.
 
 _
     },
-    examples => [
-        {
-            args => {
-                getopt_spec => {'help|h'=>sub{}, 'arg1|a=s'=>sub{},
-                                'arg2|b=s'=>sub{}},
-                words       => [qw//],
-                cword       => 0,
-            },
-            result  => ['--arg1', '--arg2', '--help', '-a', '-b', '-h'],
-        },
-        # not yet implemented
-        #{
-        #    args => {
-        #        getopt_spec => {'help|h'=>sub{}, 'arg1|a=s'=>sub{},
-        #                        'arg2|b=s'=>sub{}},
-        #        words       => [qw/--arg1 one --a/],
-        #        cword       => 2,
-        #    },
-        #    result  => ['--arg2'],
-        #    summary => '--arg1 by default is no longer completed, unless '.
-        #        'it assigns to array or has a "+" specifier',
-        #},
-    ],
 };
 sub complete_cli_arg {
     require Complete::Util;
@@ -137,87 +141,137 @@ sub complete_cli_arg {
     for my $ospec (keys %$gospec) {
         my $res = Getopt::Long::Util::parse_getopt_long_opt_spec($ospec)
             or die "Can't parse option spec '$ospec'";
+        $res->{min_vals} //= $res->{type} ? 1 : 0;
+        $res->{max_vals} //= $res->{type} || $res->{opttype} ? 1:0;
         for my $o0 (@{ $res->{opts} }) {
             my @o = $res->{is_neg} ? ($o0, "no$o0", "no-$o0") : ($o0);
             for my $o (@o) {
                 my $k = length($o) > 1 ? "--$o" : "-$o";
                 $opts{$k} = {
-                    ospec => $ospec,
-                    min_vals => $res->{min_vals} //
-                        ($res->{type} ? 1 : 0),
-                    max_vals => $res->{max_vals} //
-                        ($res->{type} || $res->{opttype} ? 1:0),
+                    name => $k,
+                    ospec => $ospec, # key to getopt specification
+                    parsed => $res,
                 };
             }
         }
     }
     my @optnames = sort keys %opts;
 
-    my $opt;
-    my $what = 'optname,arg';
-    my $lastopt;
-    my $ohash;
-    my $i = 0;
-    my $word;
-  OPT:
-    while ($i < $cword) {
-        my $w = $words->[$i] // '';
-        #say "D:i=$i, w=$w, cword=$cword, remaining=[".join(", ", @{$words}[$i+1..@$words-1])."]";
-        if ($w eq '--') {
-            $what = 'arg';
-            last OPT;
-        }
-        # should always be the case
-        if ($what =~ /optname/) {
-            # --foo^=, --foo=^, or --foo= ^
-            if ($i+1 <= @$words && ($words->[$i+1] // '') eq '=') {
-                if ($i+2 >= $cword) {
-                    if ($i+1 == $cword) {
-                        $word = '';
-                    }
-                    $ohash = $opts{$w};
-                    return [] unless $ohash;
-                    $what = 'optval';
-                    last OPT;
-                }
-                $i+=2;
-                next;
-            } elsif ($opts{$w}) {
-                $ohash = $opts{$w};
-                $what = 'optval';
-                $i += $ohash->{min_vals};
-                last OPT if $i >= $cword;
-                $what = 'optname,optval,arg';
-                for my $j (($ohash->{min_vals}+1) .. ($ohash->{max_vals})) {
-                    $i++;
-                    last OPT if $i >= @$words;
-                    if ($words->[$i] =~ /\A-/) {
-                        $what = 'optname';
-                        next OPT;
-                    }
-                }
-                $what = +($words->[$i] // '') =~ /\A-/ ?
-                    'optname' : 'optname,arg';
+    my %seen_opts;
+
+    # for each word, we try to find out whether it's supposed to complete option
+    # name, or option value, or argument, or separator (or more than one of
+    # them). plus some other information.
+    my @expects;
+
+    my $i = -1;
+
+  WORD:
+    while (1) {
+        $i++;
+        last WORD if $i >= @$words;
+        my $word = $words->[$i];
+        #say "D:i=$i, word=$word, ~~\@\$words=",~~@$words;
+
+        if ($word eq '--' && $i != $cword) {
+            $expects[$i] = {separator=>1};
+            while (1) {
                 $i++;
-                next OPT;
-            } else {
-                # assume nonexisting option to be with no value
-                $i++;
-                next OPT;
+                last WORD if $i >= @$words;
+                $expects[$i] = {arg=>1};
             }
         }
+
+        if ($word =~ /\A-/) {
+            my $opt = $word; $opt =~ s/=.*//;
+            my $opthash = _expand1($opt, \%opts);
+
+            if ($opthash) {
+                # a known argument
+                $opt = $opthash->{name};
+                $expects[$i]{optname} = $opt;
+                _mark_seen(\%seen_opts, $opt, \%opts);
+
+                my $min_vals = $opthash->{parsed}{min_vals};
+                my $max_vals = $opthash->{parsed}{max_vals};
+                #say "D:min_vals=$min_vals, max_vals=$max_vals";
+
+                # detect = after --opt
+                if ($i+1 < @$words && $words->[$i+1] eq '=') {
+                    $i++;
+                    $expects[$i] = {separator=>1, optval=>$opt, word=>''};
+                    # force a value due to =
+                    if (!$max_vals) { $min_vals = $max_vals = 1 }
+                }
+
+                for (1 .. $min_vals) {
+                    $i++;
+                    last WORD if $i >= @$words;
+                    $expects[$i]{optval} = $opt;
+                }
+                for (1 .. $max_vals-$min_vals) {
+                    last if $i+$_ >= @$words;
+                    last if $words->[$i+$_] =~ /\A-/; # a new option
+                    $expects[$i+$_]{optval} = $opt; # but can also be optname
+                }
+            } else {
+                # an unknown option, assume it doesn't require argument, unless
+                # it's --opt= or --opt=foo
+                $opt = undef;
+                $expects[$i]{optname} = $opt;
+
+                # detect = after --opt
+                if ($i+1 < @$words && $words->[$i+1] eq '=') {
+                    $i++;
+                    $expects[$i] = {separator=>1, optval=>undef, word=>''};
+                    if ($i+1 < @$words) {
+                        $i++;
+                        $expects[$i]{optval} = $opt;
+                    }
+                }
+            }
+        } else {
+            $expects[$i]{optname} = '';
+            $expects[$i]{arg} = 1;
+        }
     }
 
+    #use DD; dd \@expects;
+    #use DD; dd \%seen_opts;
+
+    my $exp = $expects[$cword];
+    my $word = $exp->{word} // $words->[$cword];
     my @res;
-    $word //= $words->[$cword] // '';
-    #say "D:---";
-    #say "D:word=$word, what=$what";
-    if ($what =~ /optname/) {
+    if (exists $exp->{optname}) {
+        my $opt = $exp->{optname};
+        # complete option names
+        my @o;
+        for (@optnames) {
+            #say "D:$_";
+            my $repeatable = 0;
+            if ($seen_opts{$_}) {
+                my $opthash = $opts{$_};
+                my $ospecval = $gospec->{$opthash->{ospec}};
+                my $parsed = $opthash->{parsed};
+                if (ref($ospecval) eq 'ARRAY') {
+                    $repeatable = 1;
+                } elsif ($parsed->{desttype} || $parsed->{is_inc}) {
+                    $repeatable = 1;
+                }
+            }
+            # skip options that have been specified and not repeatable
+            #use DD; dd {'$_'=>$_, seen=>$seen_opts{$_}, repeatable=>$repeatable, opt=>$opt};
+            next if $seen_opts{$_} && !$repeatable && (!$opt || $opt ne $_);
+            push @o, $_;
+        }
+        #use DD; dd \@o;
         push @res, @{ Complete::Util::complete_array_elem(
-            array => \@optnames, word => $word) };
+            array => \@o, word => $word) };
     }
-    if ($what =~ /optval/ && $ohash && $comps) {
-        my $comp = $comps->{$ohash->{ospec}};
+    if (exists($exp->{optval}) && $comps) {
+        my $opt = $exp->{optval};
+        my $opthash = $opts{$opt} if $opt;
+        my $comp = $comps->{$opthash->{ospec}} if $opthash;
         if (ref($comp) eq 'ARRAY') {
             push @res, @{ Complete::Util::complete_array_elem(
                 array => \@$comp, word => $word) };
@@ -231,7 +285,8 @@ sub complete_cli_arg {
             }
         }
     }
-    if ($what =~ /arg/ && $comps) {
+
+    if (exists($exp->{arg}) && $comps) {
         my $comp = $comps->{''};
         if (ref($comp) eq 'ARRAY') {
             push @res, @$comp;
@@ -264,7 +319,7 @@ Complete::Getopt::Long - Complete command-line argument using Getopt::Long speci
 
 =head1 VERSION
 
-This document describes version 0.03 of Complete::Getopt::Long (from Perl distribution Complete-Getopt-Long), released on 2014-07-24.
+This document describes version 0.04 of Complete::Getopt::Long (from Perl distribution Complete-Getopt-Long), released on 2014-07-26.
 
 =head1 SYNOPSIS
 
@@ -279,22 +334,6 @@ See L<Getopt::Long::Complete> for an easy way to use this module.
 
 Complete command-line argument using Getopt::Long specification.
 
-Examples:
-
- complete_cli_arg(
-   cword => 0,
-   getopt_spec => {
-     "arg1|a=s" => sub { ... },
-     "arg2|b=s" => sub { ... },
-     "help|h"   => sub { ... },
-   },
-   words => []
- );
-
-
-Result: C<< ["--arg1", "--arg2", "--help", "-a", "-b", "-h"] >>.
-
-
 This routine can complete option names, where the option names are retrieved
 from C<Getopt::Long> specification. If you provide completion hints in C<hints>,
 you can also complete option I<values> and I<arguments>.
@@ -341,7 +380,7 @@ Getopt::Long specification.
 
 =item * B<words>* => I<array>
 
-Command line, already broken into words.
+Command line arguments, like @ARGV.
 
 See function C<parse_cmdline> in C<Complete::Bash> on how to produce this (if
 you're using bash).
